from collections import abc

import pytest
from myclasses import MyCallable, MyIterable
from myclasses import __MyFunctor as MyFunctor


@pytest.mark.parametrize('args', [(), ([],), ((1,),), ((0, 1),), (range(1),)])
@pytest.mark.parametrize('kwargs', [{}])
def test_my_iterable(args, kwargs):
    print(MyIterable, args, kwargs)
    assert issubclass(MyIterable, abc.Iterable)
    assert not issubclass(MyIterable, abc.Container)
    assert not issubclass(MyIterable, abc.Hashable)
    assert not issubclass(MyIterable, abc.Iterator)
    assert not issubclass(MyIterable, abc.Sized)
    assert not issubclass(MyIterable, abc.Callable)
    assert not issubclass(MyIterable, abc.Sequence)
    assert not issubclass(MyIterable, abc.Set)
    assert not issubclass(MyIterable, abc.Mapping)
    try:
        assert not issubclass(MyIterable, abc.Collection)
    except AttributeError:
        pass  # Python 3.5 Compatibility
    my_iterable = MyIterable(*args, **kwargs)
    print(my_iterable.contents)
    assert isinstance(my_iterable, abc.Iterable)
    for _ in my_iterable:
        pass
    [x for x in my_iterable]

    assert my_iterable.contents == MyIterable.Type(*args, **kwargs)
    assert not my_iterable == MyIterable.Type(*args, **kwargs)
    assert my_iterable == MyIterable(*args, **kwargs)

    assert MyIterable() + [0] == MyIterable([0])
    assert MyIterable() + (0, 1) == MyIterable((0, 1))


@pytest.mark.parametrize('func, args, kwargs', [
    (MyCallable(lambda: None), (), {}),
    (MyFunctor(), (5,), {}),
])
def test_my_callable(func, args, kwargs):
    assert issubclass(MyCallable, abc.Callable)
    assert not issubclass(MyCallable, abc.Container)
    assert not issubclass(MyCallable, abc.Hashable)
    assert not issubclass(MyCallable, abc.Iterable)
    assert not issubclass(MyCallable, abc.Iterator)
    assert not issubclass(MyCallable, abc.Sized)
    assert not issubclass(MyCallable, abc.Sequence)
    assert not issubclass(MyCallable, abc.Set)
    assert not issubclass(MyCallable, abc.Mapping)
    try:
        assert not issubclass(MyCallable, abc.Collection)
    except AttributeError:
        pass  # Python 3.5 Compatibility
    my_callable = MyCallable(func)
    assert isinstance(my_callable, abc.Callable)
    try:
        func = getattr(my_callable.func, 'deepcopy', my_callable.func.copy)()
    except AttributeError:
        func = my_callable.func

    print(func)
    print(type(func))
    assert func(*args, **kwargs) == my_callable(*args, **kwargs)
    assert not my_callable.func == my_callable
    assert MyCallable(my_callable.func) == my_callable


@pytest.mark.parametrize('args, kwargs', [
    ((5,), {'y_': 3}),
])
def test_my_functor(args, kwargs):
    assert issubclass(MyFunctor, abc.Callable)
    assert not issubclass(MyFunctor, abc.Container)
    assert not issubclass(MyFunctor, abc.Hashable)
    assert not issubclass(MyFunctor, abc.Iterable)
    assert not issubclass(MyFunctor, abc.Iterator)
    assert not issubclass(MyFunctor, abc.Sized)
    assert not issubclass(MyFunctor, abc.Sequence)
    assert not issubclass(MyFunctor, abc.Set)
    assert not issubclass(MyFunctor, abc.Mapping)
    try:
        assert not issubclass(MyFunctor, abc.Collection)
    except AttributeError:
        pass  # Python 3.5 Compatibility
    my_func = MyFunctor()
    assert isinstance(my_func, abc.Callable)
    assert my_func == MyFunctor()
    assert not my_func == (lambda: None)



"""
    sphinx-autodoc-defaultargs
    ~~~~~~~~~~~~~~~~~~~~~~~~~~
    Automatic generation of default arguments
    for the Sphinx autodoc extension.
    :copyright: Copyright 2021 by Zhi Wang.
    :license: MIT, see LICENSE for details.
"""

import inspect
import itertools
import re
import sys
from typing import Any, AnyStr, Optional, Union

if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    from collections.abc import Callable, Collection, Iterable, Sequence
    List = list
    OrderedDictType = dict
    Tuple = tuple
else:
    # typing.OrderedDict is new in Python 3.7.2.
    # but OrderedDict is a subclass of dict
    from typing import Callable
    from typing import Dict as OrderedDictType
    from typing import Iterable, List, Sequence, Tuple

    try:
        from typing import Collection
    except ImportError:
        from typing import Container, Sized, TypeVar
        T_co = TypeVar('T_co', covariant=True)

        class Collection(Sized, Iterable[T_co], Container[T_co]):
            __slots__ = ()

if sys.version_info.major == 3 and sys.version_info.minor >= 7:
    OrderedDict = dict
    from re import Pattern
else:
    from collections import OrderedDict
    Pattern = type(re.compile(''))

from sphinx.application import Sphinx
from sphinx.util import logging
from sphinx.util.inspect import object_description
from sphinx.util.inspect import signature as Signature
from sphinx.util.inspect import unwrap_all

logger = logging.getLogger(__name__)

# NOTE sphinx.ext.napoleon does not support 'key' field
kw_fields = ('key', 'keyword')
param_fields = ('param', 'parameter', 'arg', 'argument') + kw_fields
type_fields = ('type', 'kwtype')
other_fields = (
    'raises', 'raise', 'except', 'exception', 'var', 'ivar', 'cvar',
    'vartype', 'returns', 'return', 'rtype', 'meta',
)


def match_field(lines: Iterable[AnyStr],
                searchfor: Union[AnyStr, Iterable[AnyStr], Pattern],
                include_blank: bool = False,
                ) -> Tuple[bool, int, int,
                           Optional[AnyStr], Optional[List[AnyStr]]]:
    """Find fields in ``lines``.

    Args:
        include_blank:
            If False, match ends as soon as an empty line is reached.
            Otherwise, it will keep searching until an nonempty line
            starting with nonwhitespace is reached, or till the end.

    Returns:
        ``(found, starting_line_index, ending_line_index,
        matched, text)``.
        Line index is left inclusive and right exclusive.
    """

    found = False
    starting_line_index = None
    ending_line_index = None
    matched = None
    text = None
    for i, line in enumerate(lines):
        # Only match once
        if not found:
            if isinstance(searchfor, Pattern):
                # Match the **beginning** of ``line``
                match = searchfor.match(line)
                if match:
                    found = True
                    starting_line_index = i
                    matched = match.group(0)
            elif isinstance(searchfor, str) or isinstance(searchfor, bytes):
                if line.startswith(searchfor):
                    found = True
                    starting_line_index = i
                    matched = searchfor
            else:
                for search_string in searchfor:
                    if line.startswith(search_string):
                        found = True
                        starting_line_index = i
                        matched = search_string
                        break

        # Found the next item
        # NOTE bytes[0] is not bytes
        #      type(line)(line[0]).isspace() <- bugged
        #      bytes(int) -> \x00 array of length arg
        elif line and line.lstrip() == line or not line and not include_blank:
            ending_line_index = i
            break

    # i is len(lines) - 1 if the loop goes completely
    if ending_line_index is None:
        if not found:
            assert starting_line_index is None
            starting_line_index = i + 1  # Set no matter found or not
        ending_line_index = i + 1  # Set no matter found or not

    assert (starting_line_index == ending_line_index) == (not found)

    # matched is `:...:`
    # should remove `:...: `

    if found:
        text = [line[len(matched) + 1:] for line in itertools.islice(
            lines, starting_line_index, ending_line_index)]

    return found, starting_line_index, ending_line_index, matched, text

# TODO
#   test default after a Note block


def rstrip_min(string: AnyStr, min_len: int,
               chars: Optional[AnyStr] = None) -> AnyStr:
    """:meth:`str.rstrip` but keep first ``min_len`` chars unchanged."""
    str_copy = string
    result = string.rstrip(chars)
    return str_copy[:min_len] if len(result) < min_len else result


def rfind_substring_in_paragraph(lines: Iterable[AnyStr],
                                 substr: AnyStr, strip: bool = True,
                                 multiline_matching: bool = False
                                 ) -> Tuple[bool, Optional[bool],
                                            Optional[Tuple[int, int]],
                                            Optional[Tuple[int, int]]]:
    """
    Find the last matching ``substr`` in ``lines``.

    Returns:
        ``(found, is_end, match_start, match_end)``.

        If not None, ``match_start`` and ``match_end`` are in format
        ``(line index, column index)``.
        Both line indices are inclusive,
        but column follows left inclusive,
        right exclusive convention.
    """

    found = False
    is_end = None
    match_start = None
    match_end = None

    # test empty substr
    if strip:
        substr = substr.strip()

    last_isempty = True
    if not multiline_matching:
        for i, line in reversed(list(enumerate(lines))):
            if last_isempty:
                last_nonempty = i
            last_isempty = False
            idx_start = line.rfind(substr)
            if idx_start >= 0:
                # found substr
                found = True
                idx_end = idx_start + len(substr)
                # Empty string is not space
                line_end = line[idx_end:].strip() if strip else line[idx_end:]
                is_end = i == last_nonempty and not line_end
                match_start = (i, idx_start)
                match_end = (i, idx_end)
                break
            elif strip and not line.strip():
                last_isempty = True
    else:
        raise NotImplementedError

    return found, is_end, match_start, match_end


def get_args(func: Callable, for_sphinx: bool = True) -> List[str]:
    signature = Signature(unwrap_all(func))

    result = []
    for name, param in signature.parameters.items():
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            name = (r'\*' if for_sphinx else '*') + name
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            name = (r'\*\*' if for_sphinx else '**') + name
        result.append(name[:-1] + r'\_'
                      if for_sphinx and name.endswith('_') else name)
    return result


def get_default_args(func: Callable,
                     for_sphinx: bool = True) -> OrderedDictType[str, Any]:
    signature = Signature(func)

    # Backward Compatibility
    #   The built-in Parameter object is guaranteed
    #   an ordered mapping in >= 3.5.
    default_args = OrderedDict()
    for k, v in signature.parameters.items():
        if v.default is not inspect.Parameter.empty:
            if for_sphinx and k.endswith('_'):
                k = '{}\\_'.format(k[:-1])
            default_args[k] = (
                v.default, v.kind == inspect.Parameter.KEYWORD_ONLY)
    return default_args


def find_arg(
        lines: Collection[str], args: Sequence[str],
        arg: str, incr: int, template=r':\S+ [\*\\]*{}:',
) -> Tuple[Optional[int], Optional[str]]:
    """Finds the argument following ``arg`` before other fields.

    Args:
        incr:
            If it is 1, find the next arg.
            If it is 0, fin the current arg.
            Should be nonnegative.

    Returns:
        (line index of the next found argument,
        the type of next argument if found)
    """

    if arg not in args:
        return None, None

    nextarg_idx = args.index(arg) + incr

    for nextarg in args[nextarg_idx:]:
        # If not lstrip, \ must be replaced by \\, because
        # the regex pattern will treat \ as an escape indicator.
        nextarg = nextarg.lstrip(r'\*')
        found, start, _, matched = match_field(
            lines, re.compile(template.format(nextarg)))[:4]
        if found:
            return start, matched.split(' ')[0][1:]

    # Assume Keyword is after Parameters
    # Other fields might come before param field
    prev_line_idx = 0
    for prevarg in reversed(args[:nextarg_idx]):
        prevarg = prevarg.lstrip(r'\*')
        found, start = match_field(
            lines, re.compile(template.format(prevarg)))[:2]
        if found:
            prev_line_idx = start
            break

    start = match_field(
        lines[prev_line_idx:], [':{}'.format(field)
                                for field in other_fields])[1]

    # Should return len(lines) if nextarg not found
    return start + prev_line_idx, None


def find_next_arg(*args, **kwargs):
    return find_arg(*args, incr=1, **kwargs)


def find_curr_arg(*args, **kwargs):
    return find_arg(*args, incr=0, **kwargs)


def process_docstring(app: Sphinx, what, name, obj, options, lines):
    """Process docstring after Sphinx.

    See `autodoc-process-docstring <https://www.sphinx-doc.org/en/master/
    usage/extensions/autodoc.html#event-autodoc-process-docstring>`_
    """

    # original_obj = obj
    if isinstance(obj, property):
        obj = obj.fget

    if not callable(obj):
        return

    if inspect.isclass(obj):
        obj = getattr(obj, '__init__', getattr(obj, '__new__', None))
        # obj = getattr(obj, '__init__')

    obj = inspect.unwrap(obj)

    default_args = get_default_args(obj)
    for argname, (default, is_keyword_only) in default_args.items():

        # what if default has \
        default = ':code:`{}`'.format(object_description(default))

        # TODO
        # should be arguments
        strip = app.config.docstring_default_arg_strip_matching
        docstring_default_arg_parenthesis = False

        # Search for parameters
        # TODO Test case: empty param
        searchfor = [':{} {}:'.format(field, argname)
                     for field in param_fields]
        param_found, param_start, param_end, param_matched, param_text = \
            match_field(lines, searchfor, include_blank=app.config.
                        docstring_default_arg_after_directives)

        if param_found:

            if app.config.docstring_default_arg_substitution not in ' '.join(
                    param_text):

                # Extracts all the flags
                for head, tail in app.config.docstring_default_arg_flags:
                    tail_found, is_end, t_start, _ = \
                        rfind_substring_in_paragraph(
                            param_text, tail, strip,
                            app.config.
                            docstring_default_arg_flags_multiline_matching)
                    if tail_found and is_end:
                        head_found, _, h_start, h_end = \
                            rfind_substring_in_paragraph(
                                param_text, head, strip,
                                app.config.
                                docstring_default_arg_flags_multiline_matching)
                        if head_found:
                            # what if default has \
                            if h_end[0] == t_start[0]:
                                default = param_text[h_end[0]
                                                     ][h_end[1]:t_start[1]]
                            else:
                                default = ' '.join(
                                    [param_text[h_end[0]][h_end[1]:]] +
                                    param_text[h_end[0] + 1:t_start[0]] +
                                    [param_text[t_start[0]][:t_start[1]]])
                            if strip:
                                default = default.strip()
                            lines[param_start + h_start[0]] = \
                                lines[param_start + h_start[0]
                                      ][:len(param_matched) + 1 + h_start[1]]
                            del lines[param_start +
                                      h_start[0] + 1: param_end]
                            param_end = param_start + h_start[0] + 1
                            break

                if strip:
                    lines[param_end - 1] = rstrip_min(
                        lines[param_end - 1], len(param_matched) + 1)

                if docstring_default_arg_parenthesis:
                    raise NotImplementedError
                else:
                    # To prevent insertion into Note directives or so
                    lines.insert(
                        param_end,
                        ' ' * len(param_matched) + ' {} {}'.format(
                            app.config.docstring_default_arg_substitution,
                            default))
        elif app.config.always_document_default_args:

            # Since ``kwargs`` (no default args) might come
            # after ``argname``, it will not be in ``default_args``.
            # Need to generate the full args list.
            next_start, next_type = find_next_arg(
                lines, get_args(obj), argname)

            if docstring_default_arg_parenthesis:
                raise NotImplementedError
            else:
                lines.insert(
                    next_start, ':{} {}: {} {}'.format(
                        'keyword' if is_keyword_only and (
                            next_type is None or
                            next_type in kw_fields) else 'param',
                        argname,
                        app.config.docstring_default_arg_substitution,
                        default))

        # Search for type
        type_found, type_start, type_end, type_matched, type_text = \
            match_field(lines, [':{} {}:'.format(field, argname)
                                for field in type_fields], include_blank=False)

        if type_found:
            type_text = ' '.join(type_text)
            if strip:
                type_text = type_text.rstrip()
                lines[type_end - 1] = rstrip_min(lines[type_end - 1],
                                                 len(type_matched) + 1)
            if not type_text.endswith('optional'):
                if not type_text.strip():
                    lines[type_start] = '{} optional'.format(type_matched)
                elif '`' in type_text:
                    # TODO check \` escape
                    lines[type_end - 1] += ', *optional*'
                else:
                    # Do not insert newline to prevent whitespace before ','
                    lines[type_end - 1] += ', optional'
        elif param_found or app.config.always_document_default_args:
            # insert type before param
            param_start, param_type = find_curr_arg(
                lines, get_args(obj), argname)
            assert any(lines[param_start].startswith(search)
                       for search in searchfor)
            lines.insert(
                param_start, ':{}type {}: optional'.format(
                    'kw' if param_type in kw_fields else '', argname))


def setup(app: Sphinx):
    app.add_config_value('always_document_default_args', False, 'html')
    app.add_config_value('docstring_default_arg_flags',
                         [('(Default: ', ')')], 'html')
    app.add_config_value(
        'docstring_default_arg_flags_multiline_matching',
        False, 'html')
    app.add_config_value(
        'docstring_default_arg_strip_matching',
        True, 'html')
    app.add_config_value(
        'docstring_default_arg_after_directives',
        False, 'html')
    app.add_config_value('docstring_default_arg_substitution',
                         '|default|', 'html')
    # app.add_config_value('docstring_default_arg_parenthesis', True, 'html')

    app.connect('autodoc-process-docstring', process_docstring)
    return dict(parallel_read_safe=True)



import os
import pathlib
import re
import shutil
import sys

import pytest
from sphinx.testing.path import path
from sphobjinv import Inventory

pytest_plugins = 'sphinx.testing.fixtures'
collect_ignore = ['roots']


@pytest.fixture(scope='session')
def inv(pytestconfig):
    cache_path = 'python{v.major}.{v.minor}/objects.inv'.format(
        v=sys.version_info)
    inv_dict = pytestconfig.cache.get(cache_path, None)
    if inv_dict is not None:
        return Inventory(inv_dict)

    print("Downloading objects.inv")
    url = 'https://docs.python.org/{v.major}.{v.minor}/objects.inv'.format(
        v=sys.version_info)
    inv = Inventory(url=url)
    pytestconfig.cache.set(cache_path, inv.json_dict())
    return inv


@pytest.fixture(autouse=True)
def remove_sphinx_projects(sphinx_test_tempdir):
    # Remove any directory which appears to be a Sphinx project from
    # the temporary directory area.
    # See https://github.com/sphinx-doc/sphinx/issues/4040
    roots_path = pathlib.Path(sphinx_test_tempdir)
    for entry in roots_path.iterdir():
        try:
            if entry.is_dir() and pathlib.Path(entry, '_build').exists():
                shutil.rmtree(str(entry))
        except PermissionError:
            pass


@pytest.fixture
def rootdir():
    return path(os.path.dirname(__file__) or '.').abspath() / 'roots'


def pytest_ignore_collect(path, config):
    version_re = re.compile(r'_py(\d)(\d)\.py$')
    match = version_re.search(path.basename)
    if match:
        version = tuple(int(x) for x in match.groups())
        if sys.version_info < version:
            return True



import itertools
import pathlib
import re
import sys
import textwrap

import pytest
from myclasses import MyCallable, MyIterable
from myclasses import __MyFunctor as MyFunctor

from sphinx_autodoc_defaultargs import (
    get_args, match_field, rfind_substring_in_paragraph, rstrip_min)


@pytest.mark.parametrize('encoding', ['utf-8'])
@pytest.mark.parametrize('args, result', [
    ([':param x:'],
        [True, 0, 1, ':param x:', '']),
    ([MyIterable((':param y:', ':parameter y:'))],
        [True, 1, 3, ':parameter y:', 'foo\nine']),
    ([':arg a:'],
        [True, 3, 4, ':arg a:', 'bar']),
    ([MyIterable((':argument b:', ':arg b:'))],
        [True, 5, 7, ':argument b:', '\n']),
    ([MyIterable((':argument b:', ':arg b:')), True],
        [True, 5, 8, ':argument b:', '\n\n']),
    ([MyIterable((':argument z:', ':arg z:'))],
        [False, 10, 10] + [None] * 2),
    ([':rtype:'],
        [True, 8, 9, ':rtype:', 'int']),
])
def test_match_field(encoding, args, result):
    lines = MyIterable(textwrap.dedent('''\
    :param x:
    :parameter y: foo
                 line
    :arg a: bar
    .. {blank}
    :argument b:
    {blank}

    :rtype: int
    ''').format(blank=' \t\v').split('\n'))

    if result[0]:
        result[-1] = result[-1].split('\n')

    assert match_field(lines, *args) == tuple(result)

    args0 = args[0]

    try:
        args[0] = re.compile(args[0])
        assert match_field(lines, *args) == tuple(result)
    except TypeError:
        pass

    if result[0]:
        result[3] = result[3].encode(encoding)
        result[4] = [r.encode(encoding) for r in result[4]]
    lines = MyIterable(line.encode(encoding) for line in lines)

    try:
        args[0] = args0.encode(encoding)
        assert match_field(lines, *args) == tuple(result)
    except AttributeError:
        args[0] = MyIterable(x.encode(encoding) for x in args0)
        assert match_field(lines, *args) == tuple(result)


# The length of 'utf-16' bytearrays doubles from the normal ones
@pytest.mark.parametrize('encoding', ['utf-8'])
@pytest.mark.parametrize('string, min_len, result, default_result', [
    ('', 0, '', ''),
    ('', 1, '', ''),
    (' \t', 1, ' \t', ' '),
    (' \t', 2, ' \t', ' \t'),
    (' a ', 1, ' ', ' a'),
    (' a ', 2, ' a', ' a'),
    (' a ', 3, ' a ', ' a '),
    (chr(0x0bf2) + ' ', 0, '', chr(0x0bf2)),
    (chr(0x0bf2) + ' ', 1, chr(0x0bf2), chr(0x0bf2)),
    (chr(0x0bf2) + ' ', 2, chr(0x0bf2) + ' ', chr(0x0bf2) + ' '),
])
def test_rstrip_min(encoding, string, min_len, result, default_result):
    chars = ' a' + chr(0x0bf2)

    assert rstrip_min(string, min_len, chars) == result
    assert rstrip_min(string, min_len) == default_result

    bstring = string.encode(encoding)
    bchars = chars.encode(encoding)
    bresult = result.encode(encoding)
    bdefault_result = default_result.encode(encoding)

    if string and string[0] == chr(0x0bf2):
        bresult = bstring[:min_len]
        bdefault_result = chr(0x0bf2).encode(encoding)

    assert rstrip_min(bstring, min_len, bchars) == bresult
    assert rstrip_min(bstring, min_len) == bdefault_result


@pytest.mark.parametrize('encoding', ['utf-8'])
@pytest.mark.parametrize('substr, strip, result', [
    (')', False, [True, False, (1, 2), (1, 3)]),
    (')', True, [True, True, (1, 2), (1, 3)]),
    (') ', False, [True, False, (1, 2), (1, 4)]),
    (') ', True, [True, True, (1, 2), (1, 3)]),
    (' )', False, [False] + [None] * 3),
    (' )', True, [True, True, (1, 2), (1, 3)]),
    ('(', False, [True, False, (1, 0), (1, 1)]),
    ('(', True, [True, False, (1, 0), (1, 1)]),
    ('( ', False, [False] + [None] * 3),
    ('( ', True, [True, False, (1, 0), (1, 1)]),
    (' (', False, [True, False, (0, 0), (0, 2)]),
    (' (', True, [True, False, (1, 0), (1, 1)]),
])
def test_rfind_substring_in_paragraph(encoding, substr, strip, result):
    lines = MyIterable(textwrap.dedent('''\
     ({blank}
    (0) {blank}
    ''').format(blank='\t\v').split('\n'))

    def test(tail):
        assert tuple(result) == rfind_substring_in_paragraph(
            lines, substr, strip)

        new_result = result.copy()
        if result[0]:
            new_result[1] = result[1] and strip
        assert tuple(new_result) == rfind_substring_in_paragraph(
            lines + [tail], substr, strip)

    test('\n')

    lines = MyIterable(line.encode(encoding) for line in lines)
    substr = substr.encode(encoding)

    test('\n'.encode(encoding))


def my_func(*, y, **kwargs_):
    pass

# TODO partial


@pytest.mark.parametrize('func, for_sphinx, result', [
    (lambda: 0, False, []),
    (lambda: 0, True, []),
    (lambda x_, *args_: 0, False, ['x_', '*args_']),
    (lambda x_, *args_: 0, True, [r'x\_', r'\*args\_']),
    (my_func, False, ['y', '**kwargs_']),
    (my_func, True, ['y', r'\*\*kwargs\_']),
    (MyFunctor(), False, ['x', 'y_']),
    (MyFunctor(), True, ['x', r'y\_']),
])
def test_get_args(func, for_sphinx, result):
    assert get_args(func, for_sphinx) == result
    # TODO
    assert get_args(MyCallable(func)) == [r'\*args', r'\*\*kwargs']


@pytest.mark.parametrize('always_document_default_args', [True, False])
@pytest.mark.sphinx('text', testroot='dummy')
def test_sphinx_output(app, status, warning, always_document_default_args):
    test_path = pathlib.Path(__file__).parent

    # Add test directory to sys.path to allow imports of dummy module.
    if str(test_path) not in sys.path:
        sys.path.insert(0, str(test_path))

    app.config.always_document_default_args = always_document_default_args
    # app.config.autodoc_mock_imports = ['mailbox']
    app.build()

    assert 'build succeeded' in status.getvalue()  # Build succeeded

    # There should be a warning
    warnings = warning.getvalue().strip()
    assert re.search('dummy_module.py:docstring of dummy_module.TestClassWithArgs.'
                     r'method_with_args_asterisk_not_escaped:\d+: '
                     'WARNING: Inline emphasis start-string without end-string.',
                     warnings), warnings

    pre_dict = {
        'param': '\n\n   Parameters:\n'
        '      **{var}**{type_str} --  "{default}"',
        'midparam': '\n\n      * **{var}**{type_str} --  "{default}"',
        'kw': '\n\n   Keyword Arguments:\n'
        '      **{var}**{type_str} --  "{default}"',
    }
    varlist = ('x', 'a')
    typelist = ('', 'int', 'str')
    defaultlist = ('None', '0')

    format_args = {}
    for (pre, text), var, typename, default, indentation_level in \
            itertools.product(pre_dict.items(), varlist,
                              typelist, defaultlist, range(4)):
        key = '{}_{}_t{}_v{}_ind{}'.format(pre, var, typename, default,
                                           indentation_level)
        format_args[key] = textwrap.indent(
            text.format(var=var, default=default,
                        type_str=' ({}*optional*)'.format(
                            '*{}**, *'.format(typename) if typename else ''),
                        ),
            '   ' * indentation_level) \
            if always_document_default_args else ''

    text_path = pathlib.Path(app.srcdir) / '_build' / 'text' / 'index.txt'
    with text_path.open('r') as f:
        text_contents = f.read().replace('–', '--')
        expected_contents = textwrap.dedent('''\
        Dummy Module
        ************

        dummy_module.func(x=None, y=None)

           Example docstring.

           Parameters:
              * **x** (*optional*) -- The default value "None" will be added
                here.  "None"

              * **y** (*optional*) --

                The text of default value is unchanged.

                 "'Default Value'"

        class dummy_module.TestClassWithReturn(x=None)

           Class docstring.{param_x_t_vNone_ind0}

           method_with_last_defargs(x, y, a=0, **kw)

              Method docstring.

              Returns:
                 0

              Parameters:
                 * **x** -- foo

                 * **y** -- bar{midparam_a_t_v0_ind1}

              Keyword Arguments:
                 ****kw** -- null

           method_with_mid_defargs(x, a=0, y='')

              Method docstring.

              Returns:
                 0

              Parameters:
                 * **x** (*int**, **str*) -- foo{midparam_a_t_v0_ind1}

                 * **y** (*str**, **optional*) -- bar  empty

        class dummy_module.TestClassWithArgs

           Class docstring.

           method_with_args(a, x=0, *args)

              Method docstring.

              Parameters:
                 * **a** -- foo{midparam_x_t_v0_ind1}

                 * ***args** -- bar

           method_with_args_asterisk_not_escaped(a, x=0, *args)

              Method docstring.

              Parameters:
                 * **a** -- foo{midparam_x_t_v0_ind1}

                 * ***args** --

                   bar

           method_with_args_no_asterisk(a, x=0, *args)

              Method docstring.

              Parameters:
                 * **a** -- foo{midparam_x_t_v0_ind1}

                 * **args** -- bar

           method_with_kwargs_as_kw_x_forced_param(a, *, x=0, **kwargs)

              Method docstring.

              Parameters:
                 * **a** -- foo{midparam_x_t_v0_ind1}

                 * ****kwargs** -- bar

           method_with_kwargs_as_kw_x_kw(*, a, x=0, **kwargs)

              Method docstring.

              Keyword Arguments:
                 * **a** -- foo{midparam_x_t_v0_ind1}

                 * ****kwargs** -- bar

           method_with_kwargs_as_kw_x_param(x=0, {slash}**kwargs)

              Method docstring.{param_x_t_v0_ind1}

              Keyword Arguments:
                 ****kwargs** -- foo

           method_with_kwargs_as_param(a, x=0, **kwargs)

              Method docstring.

              Parameters:
                 * **a** -- foo{midparam_x_t_v0_ind1}

                 * ****kwargs** -- bar

           method_with_last_kw_optional(*args, x=0)

              Method docstring.

              Returns:
                 0

              Parameters:
                 ***args** -- foo{kw_x_t_v0_ind1}

              Raises:
                 **ValueError** -- If "x" is nonzero.
        ''')
        print(text_contents)
        expected_contents = expected_contents.format(
            slash='' if sys.version_info[:2] < (3, 8) else '/, ',
            **format_args).replace('–', '--')
        assert text_contents == expected_contents



import sys
import textwrap


def func(x=None, y=None):
    """
    Example docstring.

    :param x: The default value ``None`` will be added here.
    :param y: The text of default value is unchanged.
              (Default: ``'Default Value'``)
    """

    if y is None:
        y = 'Default Value'
    pass


class TestClassWithReturn:
    """
    Class docstring.
    """

    def __init__(self, x=None):
        pass

    # a should be param
    def method_with_last_defargs(self, x, y, a=0, **kw):
        r"""
        Method docstring.

        :return: 0
        :param x: foo
        :param y: bar
        :keyword \*\*kw: null
        """
        return 0

    def method_with_mid_defargs(self, x, a=0, y=''):
        """
        Method docstring.

        :return: 0
        :param x: foo
        :type x: int, str
        :type y: str
        :parameter y: bar (Default: empty)
        """
        return 0


class TestClassWithArgs:
    """
    Class docstring.
    """

    def method_with_args(self, a, x=0, *args):
        r"""
        Method docstring.

        :param a: foo
        :param \*args: bar
        """
        pass

    def method_with_args_asterisk_not_escaped(self, a, x=0, *args):
        r"""
        Method docstring.

        :arg a: foo
        :arg *args: bar
        """
        pass

    def method_with_args_no_asterisk(self, a, x=0, *args):
        r"""
        Method docstring.

        :argument a: foo
        :argument args: bar
        """
        pass

    # param
    def method_with_kwargs_as_kw_x_forced_param(self, a, *, x=0, **kwargs):
        r"""
        Method docstring.

        :param a: foo
        :param \*\*kwargs: bar
        """
        pass

    # keyword
    def method_with_kwargs_as_kw_x_kw(self, *, a, x=0, **kwargs):
        r"""
        Method docstring.

        :keyword a: foo
        :keyword \*\*kwargs: bar
        """
        pass

    # param
    exec(textwrap.dedent('''\
    def method_with_kwargs_as_kw_x_param(self, x=0, {}**kwargs):
        {}
        pass'''.format(
        '' if sys.version_info[:2] < (3, 8) else '/, ', repr(r"""
            Method docstring.

            :keyword \*\*kwargs: foo
            """))))

    # param
    def method_with_kwargs_as_param(self, a, x=0, **kwargs):
        r"""
        Method docstring.

        :param a: foo
        :param \*\*kwargs: bar
        """
        pass

    def method_with_last_kw_optional(self, *args, x=0):
        r"""
        Method docstring.

        :return: 0
        :param \*args: foo
        :raises ValueError: If ``x`` is nonzero.
        """
        if x != 0:
            raise ValueError
        return x



import pathlib
import sys

# Make dummy_module.py available for autodoc.
sys.path.insert(0, str(pathlib.Path(__file__).parent))


master_doc = 'index'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_defaultargs',
]

rst_prolog = """
.. |default| raw:: html

    <div class="default-value-section">""" + \
    ' <span class="default-value-label">Default:</span>'

always_document_default_args = True

# html_theme = "sphinx_rtd_theme"
#
#
# def setup(app):
#     app.add_css_file('css/custom.css')
#     # app.add_stylesheet('css/custom.css')



import sys


class MyIterable(object):
    """My Iterable class."""

    Type = tuple

    def __init__(self, *args, **kwargs):
        self.contents = self.Type(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        return self.contents.__iter__(*args, **kwargs)

    def __eq__(self, other):
        return isinstance(
            other, type(self)) and self.contents == other.contents

    def __add__(self, other):
        try:
            other = other.contents
        except AttributeError:
            pass
        return self.__class__(self.contents + self.Type(other))


class MyCallable(object):
    """My Callable class."""

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    # must keep the same type
    def __eq__(self, other):
        return isinstance(other, type(self)) and self.func == other.func


class __MyFunctor(object):
    exec('def __call__(self=None, x=None, {}y_=0): pass'.format(
        '' if sys.version_info[:2] < (3, 8) else '/, '))

    def __eq__(self, other):
        try:
            return isinstance(other, self.__class__)
        except Exception:
            return False

    def __method(self=None):
        pass

    @classmethod
    def _classmethod_(cls=r'_\\'):
        pass

    @staticmethod
    def _staticmethod(x=0):
        pass
    __staticmethod = _staticmethod
    _classmethod = _classmethod_
